@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: MIT OR GPL-3.0-only */
 /* auth.c
 ** strophe XMPP client library -- auth functions and handlers
 **
@@ -6,7 +7,7 @@
 **  This software is provided AS-IS with no warranty, either express or
 **  implied.
 **
-** This program is dual licensed under the MIT and GPLv3 licenses.
+** This program is dual licensed under the MIT or GPLv3 licenses.
 */
 
 /** @file
@@ -61,6 +62,7 @@
 
 static void _auth(xmpp_conn_t *conn);
 static void _auth_legacy(xmpp_conn_t *conn);
+static void _handle_open_compress(xmpp_conn_t *conn);
 static void _handle_open_sasl(xmpp_conn_t *conn);
 static void _handle_open_tls(xmpp_conn_t *conn);
 
@@ -71,6 +73,9 @@ static int _handle_component_hs_response(xmpp_conn_t *conn,
 
 static int
 _handle_features_sasl(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata);
+static int _handle_features_compress(xmpp_conn_t *conn,
+                                     xmpp_stanza_t *stanza,
+                                     void *userdata);
 static int
 _handle_sasl_result(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata);
 static int _handle_digestmd5_challenge(xmpp_conn_t *conn,
@@ -82,7 +87,8 @@ static int _handle_digestmd5_rspauth(xmpp_conn_t *conn,
 static int _handle_scram_challenge(xmpp_conn_t *conn,
                                    xmpp_stanza_t *stanza,
                                    void *userdata);
-static char *_make_scram_init_msg(xmpp_conn_t *conn);
+struct scram_user_data;
+static int _make_scram_init_msg(struct scram_user_data *scram);
 
 static int _handle_missing_features_sasl(xmpp_conn_t *conn, void *userdata);
 static int _handle_missing_bind(xmpp_conn_t *conn, void *userdata);
@@ -205,12 +211,54 @@ static int _handle_missing_features(xmpp_conn_t *conn, void *userdata)
     return 0;
 }
 
+typedef void (*text_handler)(xmpp_conn_t *conn, const char *text);
+static void _foreach_child(xmpp_conn_t *conn,
+                           xmpp_stanza_t *parent,
+                           const char *name,
+                           text_handler hndl)
+{
+    xmpp_stanza_t *children;
+    for (children = xmpp_stanza_get_children(parent); children;
+         children = xmpp_stanza_get_next(children)) {
+        const char *child_name = xmpp_stanza_get_name(children);
+        if (child_name && strcmp(child_name, name) == 0) {
+            char *text = xmpp_stanza_get_text(children);
+            if (text == NULL)
+                continue;
+
+            hndl(conn, text);
+
+            strophe_free(conn->ctx, text);
+        }
+    }
+}
+
+static void _handle_sasl_children(xmpp_conn_t *conn, const char *text)
+{
+    if (strcasecmp(text, "PLAIN") == 0) {
+        conn->sasl_support |= SASL_MASK_PLAIN;
+    } else if (strcasecmp(text, "EXTERNAL") == 0 &&
+               (conn->tls_client_cert || conn->tls_client_key)) {
+        conn->sasl_support |= SASL_MASK_EXTERNAL;
+    } else if (strcasecmp(text, "DIGEST-MD5") == 0) {
+        conn->sasl_support |= SASL_MASK_DIGESTMD5;
+    } else if (strcasecmp(text, "ANONYMOUS") == 0) {
+        conn->sasl_support |= SASL_MASK_ANONYMOUS;
+    } else {
+        size_t n;
+        for (n = 0; n < scram_algs_num; ++n) {
+            if (strcasecmp(text, scram_algs[n]->scram_name) == 0) {
+                conn->sasl_support |= scram_algs[n]->mask;
+                break;
+            }
+        }
+    }
+}
+
 static int
 _handle_features(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
 {
-    xmpp_stanza_t *child, *mech;
-    const char *ns;
-    char *text;
+    xmpp_stanza_t *child;
 
     UNUSED(userdata);
 
@@ -220,10 +268,9 @@ _handle_features(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
     /* check for TLS */
     if (!conn->secured) {
         if (!conn->tls_disabled) {
-            child = xmpp_stanza_get_child_by_name(stanza, "starttls");
-            if (child) {
-                ns = xmpp_stanza_get_ns(child);
-                conn->tls_support = ns != NULL && strcmp(ns, XMPP_NS_TLS) == 0;
+            if (xmpp_stanza_get_child_by_name_and_ns(stanza, "starttls",
+                                                     XMPP_NS_TLS)) {
+                conn->tls_support = 1;
             }
         } else {
             conn->tls_support = 0;
@@ -231,36 +278,10 @@ _handle_features(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
     }
 
     /* check for SASL */
-    child = xmpp_stanza_get_child_by_name(stanza, "mechanisms");
-    ns = child ? xmpp_stanza_get_ns(child) : NULL;
-    if (child && ns && strcmp(ns, XMPP_NS_SASL) == 0) {
-        for (mech = xmpp_stanza_get_children(child); mech;
-             mech = xmpp_stanza_get_next(mech)) {
-            if (xmpp_stanza_get_name(mech) &&
-                strcmp(xmpp_stanza_get_name(mech), "mechanism") == 0) {
-                text = xmpp_stanza_get_text(mech);
-                if (text == NULL)
-                    continue;
-
-                if (strcasecmp(text, "PLAIN") == 0)
-                    conn->sasl_support |= SASL_MASK_PLAIN;
-                else if (strcasecmp(text, "EXTERNAL") == 0 &&
-                         (conn->tls_client_cert || conn->tls_client_key))
-                    conn->sasl_support |= SASL_MASK_EXTERNAL;
-                else if (strcasecmp(text, "DIGEST-MD5") == 0)
-                    conn->sasl_support |= SASL_MASK_DIGESTMD5;
-                else if (strcasecmp(text, "SCRAM-SHA-1") == 0)
-                    conn->sasl_support |= SASL_MASK_SCRAMSHA1;
-                else if (strcasecmp(text, "SCRAM-SHA-256") == 0)
-                    conn->sasl_support |= SASL_MASK_SCRAMSHA256;
-                else if (strcasecmp(text, "SCRAM-SHA-512") == 0)
-                    conn->sasl_support |= SASL_MASK_SCRAMSHA512;
-                else if (strcasecmp(text, "ANONYMOUS") == 0)
-                    conn->sasl_support |= SASL_MASK_ANONYMOUS;
-
-                strophe_free(conn->ctx, text);
-            }
-        }
+    child = xmpp_stanza_get_child_by_name_and_ns(stanza, "mechanisms",
+                                                 XMPP_NS_SASL);
+    if (child) {
+        _foreach_child(conn, child, "mechanism", _handle_sasl_children);
     }
 
     /* Disable PLAIN when other secure mechanisms are supported */
@@ -334,7 +355,9 @@ _handle_sasl_result(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
                       (char *)userdata);
 
         /* reset parser */
-        conn_prepare_reset(conn, _handle_open_sasl);
+        conn_prepare_reset(conn, conn->compression.allowed
+                                     ? _handle_open_compress
+                                     : _handle_open_sasl);
 
         /* send stream tag */
         conn_open_stream(conn);
@@ -438,7 +461,11 @@ static int _handle_digestmd5_rspauth(xmpp_conn_t *conn,
 }
 
 struct scram_user_data {
+    xmpp_conn_t *conn;
+    int sasl_plus;
     char *scram_init;
+    char *channel_binding;
+    const char *first_bare;
     const struct hash_alg *alg;
 };
 
@@ -470,8 +497,9 @@ static int _handle_scram_challenge(xmpp_conn_t *conn,
         if (!challenge)
             goto err;
 
-        response = sasl_scram(conn->ctx, scram_ctx->alg, challenge,
-                              scram_ctx->scram_init, conn->jid, conn->pass);
+        response =
+            sasl_scram(conn->ctx, scram_ctx->alg, scram_ctx->channel_binding,
+                       challenge, scram_ctx->first_bare, conn->jid, conn->pass);
         strophe_free(conn->ctx, challenge);
         if (!response)
             goto err;
@@ -505,7 +533,8 @@ static int _handle_scram_challenge(xmpp_conn_t *conn,
          */
         rc = _handle_sasl_result(conn, stanza,
                                  (void *)scram_ctx->alg->scram_name);
-        strophe_free(conn->ctx, scram_ctx->scram_init);
+        strophe_free_and_null(conn->ctx, scram_ctx->channel_binding);
+        strophe_free_and_null(conn->ctx, scram_ctx->scram_init);
         strophe_free(conn->ctx, scram_ctx);
     }
 
@@ -516,33 +545,103 @@ err_release_auth:
 err_free_response:
     strophe_free(conn->ctx, response);
 err:
-    strophe_free(conn->ctx, scram_ctx->scram_init);
+    strophe_free_and_null(conn->ctx, scram_ctx->channel_binding);
+    strophe_free_and_null(conn->ctx, scram_ctx->scram_init);
     strophe_free(conn->ctx, scram_ctx);
     disconnect_mem_error(conn);
     return 0;
 }
 
-static char *_make_scram_init_msg(xmpp_conn_t *conn)
+static int _make_scram_init_msg(struct scram_user_data *scram)
 {
+    xmpp_conn_t *conn = scram->conn;
     xmpp_ctx_t *ctx = conn->ctx;
-    size_t message_len;
-    char *node;
-    char *message;
-    char nonce[32];
+    const void *binding_data;
+    const char *binding_type;
+    char *node, *message;
+    size_t message_len, binding_type_len = 0, binding_data_len;
+    int l, is_secured = xmpp_conn_is_secured(conn);
+    /* This buffer must be able to hold:
+     * "p=<10 bytes binding type>,,<36 bytes binding data>"
+     * + alignment */
+    char buf[56];
+
+    if (scram->sasl_plus) {
+        if (!is_secured) {
+            strophe_error(
+                ctx, "xmpp",
+                "SASL: Server requested a -PLUS variant to authenticate, "
+                "but the connection is not secured. This is an error on "
+                "the server side we can't do anything about.");
+            return -1;
+        }
+        if (tls_init_channel_binding(conn->tls, &binding_type,
+                                     &binding_type_len)) {
+            return -1;
+        }
+        /* directly account for the '=' char in 'p=<binding-type>' */
+        binding_type_len += 1;
+    }
 
     node = xmpp_jid_node(ctx, conn->jid);
     if (!node) {
-        return NULL;
+        return -1;
     }
-    xmpp_rand_nonce(ctx->rand, nonce, sizeof(nonce));
-    message_len = strlen(node) + strlen(nonce) + 8 + 1;
+    /* 32 bytes nonce is enough */
+    xmpp_rand_nonce(ctx->rand, buf, 33);
+    message_len = strlen(node) + strlen(buf) + 8 + binding_type_len + 1;
     message = strophe_alloc(ctx, message_len);
-    if (message) {
-        strophe_snprintf(message, message_len, "n,,n=%s,r=%s", node, nonce);
+    if (!message) {
+        goto err_node;
     }
+    /* increase length to account for 'y,,', 'n,,' or 'p,,'.
+     * In the 'p' case the '=' sign has already been accounted for above.
+     */
+    binding_type_len += 3;
+    if (scram->sasl_plus) {
+        l = strophe_snprintf(message, message_len, "p=%s,,n=%s,r=%s",
+                             binding_type, node, buf);
+    } else {
+        l = strophe_snprintf(message, message_len, "%c,,n=%s,r=%s",
+                             is_secured ? 'y' : 'n', node, buf);
+    }
+    if (l < 0 || (size_t)l >= message_len) {
+        goto err_msg;
+    }
+    if (binding_type_len > sizeof(buf)) {
+        goto err_msg;
+    }
+    /* Make `first_bare` point to the 'n' of 'n=<node>' of the
+     * client-first-message */
+    scram->first_bare = message + binding_type_len;
+    memcpy(buf, message, binding_type_len);
+    if (scram->sasl_plus) {
+        binding_data =
+            tls_get_channel_binding_data(conn->tls, &binding_data_len);
+        if (!binding_data) {
+            goto err_msg;
+        }
+        if (binding_data_len > sizeof(buf) - binding_type_len) {
+            strophe_error(ctx, "xmpp", "Channel binding data is too long (%zu)",
+                          binding_data_len);
+            goto err_msg;
+        }
+        memcpy(&buf[binding_type_len], binding_data, binding_data_len);
+        binding_type_len += binding_data_len;
+    }
+    scram->channel_binding =
+        xmpp_base64_encode(ctx, (void *)buf, binding_type_len);
+    memset(buf, 0, binding_type_len);
     strophe_free(ctx, node);
+    scram->scram_init = message;
 
-    return message;
+    return 0;
+
+err_msg:
+    strophe_free(ctx, message);
+err_node:
+    strophe_free(ctx, node);
+    return -1;
 }
 
 static xmpp_stanza_t *_make_starttls(xmpp_conn_t *conn)
@@ -635,7 +734,7 @@ static void _auth(xmpp_conn_t *conn)
         return;
     }
 
-    if (anonjid && conn->sasl_support & SASL_MASK_ANONYMOUS) {
+    if (anonjid && (conn->sasl_support & SASL_MASK_ANONYMOUS)) {
         /* some crap here */
         auth = _make_sasl_auth(conn, "ANONYMOUS");
         if (!auth) {
@@ -701,22 +800,26 @@ static void _auth(xmpp_conn_t *conn)
             "Password hasn't been set, and SASL ANONYMOUS unsupported.");
         xmpp_disconnect(conn);
     } else if (conn->sasl_support & SASL_MASK_SCRAM) {
+        size_t n;
         scram_ctx = strophe_alloc(conn->ctx, sizeof(*scram_ctx));
-        if (conn->sasl_support & SASL_MASK_SCRAMSHA512)
-            scram_ctx->alg = &scram_sha512;
-        else if (conn->sasl_support & SASL_MASK_SCRAMSHA256)
-            scram_ctx->alg = &scram_sha256;
-        else if (conn->sasl_support & SASL_MASK_SCRAMSHA1)
-            scram_ctx->alg = &scram_sha1;
+        memset(scram_ctx, 0, sizeof(*scram_ctx));
+        for (n = 0; n < scram_algs_num; ++n) {
+            if (conn->sasl_support & scram_algs[n]->mask) {
+                scram_ctx->alg = scram_algs[n];
+                break;
+            }
+        }
+
         auth = _make_sasl_auth(conn, scram_ctx->alg->scram_name);
         if (!auth) {
             disconnect_mem_error(conn);
             return;
         }
 
-        /* don't free scram_init on success */
-        scram_ctx->scram_init = _make_scram_init_msg(conn);
-        if (!scram_ctx->scram_init) {
+        scram_ctx->conn = conn;
+        scram_ctx->sasl_plus =
+            scram_ctx->alg->mask & SASL_MASK_SCRAM_PLUS ? 1 : 0;
+        if (_make_scram_init_msg(scram_ctx)) {
             strophe_free(conn->ctx, scram_ctx);
             xmpp_stanza_release(auth);
             disconnect_mem_error(conn);
@@ -752,7 +855,7 @@ static void _auth(xmpp_conn_t *conn)
 
         send_stanza(conn, auth, XMPP_QUEUE_STROPHE);
 
-        /* SASL SCRAM-SHA-1 was tried, unset flag */
+        /* SASL algorithm was tried, unset flag */
         conn->sasl_support &= ~scram_ctx->alg->mask;
     } else if (conn->sasl_support & SASL_MASK_DIGESTMD5) {
         auth = _make_sasl_auth(conn, "DIGEST-MD5");
@@ -812,10 +915,10 @@ static void _auth(xmpp_conn_t *conn)
     }
 }
 
-static void _auth_success(xmpp_conn_t *conn)
+static void _stream_negotiation_success(xmpp_conn_t *conn)
 {
     tls_clear_password_cache(conn);
-    conn->authenticated = 1;
+    conn->stream_negotiation_completed = 1;
     /* call connection handler */
     conn->conn_handler(conn, XMPP_CONN_CONNECT, 0, NULL, conn->userdata);
 }
@@ -865,6 +968,17 @@ static void _handle_open_sasl(xmpp_conn_t *conn)
                       NULL);
 }
 
+/* called when stream:stream tag received after compression has been enabled */
+static void _handle_open_compress(xmpp_conn_t *conn)
+{
+    strophe_debug(conn->ctx, "xmpp", "Reopened stream successfully.");
+
+    /* setup stream:features handlers */
+    handler_add(conn, _handle_features_compress, XMPP_NS_STREAMS, "features",
+                NULL, NULL);
+    handler_add_timed(conn, _handle_missing_features, FEATURES_TIMEOUT, NULL);
+}
+
 static int _do_bind(xmpp_conn_t *conn, xmpp_stanza_t *bind)
 {
     xmpp_stanza_t *iq, *res, *text;
@@ -877,6 +991,7 @@ static int _do_bind(xmpp_conn_t *conn, xmpp_stanza_t *bind)
     /* send bind request */
     iq = xmpp_iq_new(conn->ctx, "set", "_xmpp_bind1");
     if (!iq) {
+        xmpp_stanza_release(bind);
         disconnect_mem_error(conn);
         return 0;
     }
@@ -921,6 +1036,62 @@ static int _do_bind(xmpp_conn_t *conn, xmpp_stanza_t *bind)
     return 0;
 }
 
+static int _handle_compress_result(xmpp_conn_t *const conn,
+                                   xmpp_stanza_t *const stanza,
+                                   void *const userdata)
+{
+    const char *name = xmpp_stanza_get_name(stanza);
+
+    UNUSED(userdata);
+
+    if (!name)
+        return 0;
+    if (strcmp(name, "compressed") == 0) {
+        /* Stream compression enabled, we need to restart the stream */
+        strophe_debug(conn->ctx, "xmpp", "Stream compression enabled");
+
+        /* reset parser */
+        conn_prepare_reset(conn, _handle_open_sasl);
+
+        /* make compression effective */
+        compression_init(conn);
+
+        /* send stream tag */
+        conn_open_stream(conn);
+    }
+    return 0;
+}
+
+static int _handle_features_compress(xmpp_conn_t *conn,
+                                     xmpp_stanza_t *stanza,
+                                     void *userdata)
+{
+    const char *compress = "<compress xmlns='" XMPP_NS_COMPRESSION
+                           "'><method>zlib</method></compress>";
+    xmpp_stanza_t *child;
+
+    /* remove missing features handler */
+    xmpp_timed_handler_delete(conn, _handle_missing_features);
+
+    /* check for compression */
+    child = xmpp_stanza_get_child_by_name_and_ns(stanza, "compression",
+                                                 XMPP_NS_FEATURE_COMPRESSION);
+    if (conn->compression.allowed && child) {
+        _foreach_child(conn, child, "method",
+                       compression_handle_feature_children);
+    }
+
+    if (conn->compression.supported) {
+        send_raw(conn, compress, strlen(compress), XMPP_QUEUE_STROPHE, NULL);
+        handler_add(conn, _handle_compress_result, XMPP_NS_COMPRESSION, NULL,
+                    NULL, NULL);
+    } else {
+        return _handle_features_sasl(conn, stanza, userdata);
+    }
+
+    return 0;
+}
+
 static int
 _handle_features_sasl(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
 {
@@ -1017,7 +1188,7 @@ static int
 _handle_bind(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
 {
     const char *type;
-    xmpp_stanza_t *iq, *enable, *session, *binding, *jid_stanza;
+    xmpp_stanza_t *iq, *session, *binding, *jid_stanza, *enable = NULL;
 
     UNUSED(userdata);
 
@@ -1040,6 +1211,23 @@ _handle_bind(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
             }
         }
 
+        /* send enable directly after the bind request */
+        if (conn->sm_state->sm_support && !conn->sm_disable) {
+            enable = xmpp_stanza_new(conn->ctx);
+            if (!enable) {
+                disconnect_mem_error(conn);
+                return 0;
+            }
+            xmpp_stanza_set_name(enable, "enable");
+            xmpp_stanza_set_ns(enable, XMPP_NS_SM);
+            if (!conn->sm_state->dont_request_resume)
+                xmpp_stanza_set_attribute(enable, "resume", "true");
+            handler_add(conn, _handle_sm, XMPP_NS_SM, NULL, NULL, NULL);
+            send_stanza(conn, enable, XMPP_QUEUE_SM_STROPHE);
+            conn->sm_state->sm_sent_nr = 0;
+            conn->sm_state->sm_enabled = 1;
+        }
+
         /* establish a session if required */
         if (conn->session_required) {
             /* setup response handlers */
@@ -1070,22 +1258,12 @@ _handle_bind(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
             send_stanza(conn, iq, XMPP_QUEUE_STROPHE);
         }
 
-        if (conn->sm_state->sm_support && !conn->sm_disable) {
-            enable = xmpp_stanza_new(conn->ctx);
-            if (!enable) {
-                disconnect_mem_error(conn);
-                return 0;
-            }
-            xmpp_stanza_set_name(enable, "enable");
-            xmpp_stanza_set_ns(enable, XMPP_NS_SM);
-            if (!conn->sm_state->dont_request_resume)
-                xmpp_stanza_set_attribute(enable, "resume", "true");
-            handler_add(conn, _handle_sm, XMPP_NS_SM, NULL, NULL, NULL);
-            send_stanza(conn, enable, XMPP_QUEUE_SM_STROPHE);
-        }
-
-        if (!conn->session_required) {
-            _auth_success(conn);
+        /* if there's no xmpp session required and we didn't try to enable
+         * stream-management, we're done here and the stream-negotiation was
+         * successful
+         */
+        if (!conn->session_required && !enable) {
+            _stream_negotiation_success(conn);
         }
     } else {
         strophe_error(conn->ctx, "xmpp", "Server sent malformed bind reply.");
@@ -1122,7 +1300,7 @@ _handle_session(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
     } else if (type && strcmp(type, "result") == 0) {
         strophe_debug(conn->ctx, "xmpp", "Session establishment successful.");
 
-        _auth_success(conn);
+        _stream_negotiation_success(conn);
     } else {
         strophe_error(conn->ctx, "xmpp",
                       "Server sent malformed session reply.");
@@ -1153,23 +1331,59 @@ static int _handle_missing_legacy(xmpp_conn_t *conn, void *userdata)
     return 0;
 }
 
+static int _get_h_attribute(xmpp_stanza_t *stanza, unsigned long *ul_h)
+{
+    const char *h = xmpp_stanza_get_attribute(stanza, "h");
+    if (!h || string_to_ul(h, ul_h)) {
+        strophe_error(
+            stanza->ctx, "xmpp",
+            "SM error: failed parsing 'h', \"%s\" got converted to %llu.",
+            STR_MAYBE_NULL(h), *ul_h);
+        return -1;
+    }
+    return 0;
+}
+
+static void _sm_queue_cleanup(xmpp_conn_t *conn, unsigned long ul_h)
+{
+    xmpp_send_queue_t *e;
+    while ((e = peek_queue_front(&conn->sm_state->sm_queue))) {
+        if (e->sm_h >= ul_h)
+            break;
+        e = pop_queue_front(&conn->sm_state->sm_queue);
+        strophe_free(conn->ctx, queue_element_free(conn->ctx, e));
+    }
+}
+
+static void _sm_queue_resend(xmpp_conn_t *conn)
+{
+    xmpp_send_queue_t *e;
+    while ((e = pop_queue_front(&conn->sm_state->sm_queue))) {
+        /* Re-send what was already sent out and is still in the
+         * SM queue (i.e. it hasn't been ACK'ed by the server)
+         */
+        strophe_debug_verbose(2, conn->ctx, "conn", "SM_Q_RESEND: %p, h=%lu", e,
+                              e->sm_h);
+        send_raw(conn, e->data, e->len, e->owner, NULL);
+        strophe_free(conn->ctx, queue_element_free(conn->ctx, e));
+    }
+}
+
 static int _handle_sm(xmpp_conn_t *const conn,
                       xmpp_stanza_t *const stanza,
                       void *const userdata)
 {
-    xmpp_stanza_t *failed_cause;
-    const char *name, *id, *previd, *resume, *h, *cause;
-    xmpp_send_queue_t *e;
+    xmpp_stanza_t *failed_cause, *bind = NULL;
+    const char *name, *id, *previd, *resume, *cause;
     unsigned long ul_h = 0;
 
     UNUSED(userdata);
 
     name = xmpp_stanza_get_name(stanza);
     if (!name)
-        goto LBL_ERR;
+        goto err_sm;
 
     if (strcmp(name, "enabled") == 0) {
-        conn->sm_state->sm_enabled = 1;
         conn->sm_state->sm_handled_nr = 0;
         resume = xmpp_stanza_get_attribute(stanza, "resume");
         if (resume && (strcasecmp(resume, "true") || strcmp(resume, "1"))) {
@@ -1179,11 +1393,17 @@ static int _handle_sm(xmpp_conn_t *const conn,
                               "SM error: server said it can resume, but "
                               "didn't provide an ID.");
                 name = NULL;
-                goto LBL_ERR;
+                goto err_sm;
             }
             conn->sm_state->can_resume = 1;
             conn->sm_state->id = strophe_strdup(conn->ctx, id);
         }
+        /* We maybe have stuff in the SM queue if we tried to resume, but the
+         * server doesn't remember all details of our session, but the `h` was
+         * still available.
+         */
+        _sm_queue_resend(conn);
+        _stream_negotiation_success(conn);
     } else if (strcmp(name, "resumed") == 0) {
         previd = xmpp_stanza_get_attribute(stanza, "previd");
         if (!previd || strcmp(previd, conn->sm_state->previd)) {
@@ -1191,16 +1411,11 @@ static int _handle_sm(xmpp_conn_t *const conn,
                           "SM error: previd didn't match, ours is \"%s\".",
                           conn->sm_state->previd);
             name = NULL;
-            goto LBL_ERR;
+            goto err_sm;
         }
-        h = xmpp_stanza_get_attribute(stanza, "h");
-        if (!h || string_to_ul(h, &ul_h)) {
-            strophe_error(conn->ctx, "xmpp",
-                          "SM error: failed parsing 'h', it got converted "
-                          "to %llu.",
-                          ul_h);
+        if (_get_h_attribute(stanza, &ul_h)) {
             name = NULL;
-            goto LBL_ERR;
+            goto err_sm;
         }
         conn->sm_state->sm_enabled = 1;
         conn->sm_state->id = conn->sm_state->previd;
@@ -1211,54 +1426,53 @@ static int _handle_sm(xmpp_conn_t *const conn,
             conn->sm_state->sm_sent_nr = conn->sm_state->sm_queue.head->sm_h;
         else
             conn->sm_state->sm_sent_nr = ul_h;
-        while ((e = pop_queue_front(&conn->sm_state->sm_queue))) {
-            if (e->sm_h >= ul_h) {
-                /* Re-send what was already sent out and is still in the
-                 * SM queue (i.e. it hasn't been ACK'ed by the server)
-                 */
-                send_raw(conn, e->data, e->len, e->owner, NULL);
-            }
-            strophe_free(conn->ctx, queue_element_free(conn->ctx, e));
-        }
+        _sm_queue_cleanup(conn, ul_h);
+        _sm_queue_resend(conn);
         strophe_debug(conn->ctx, "xmpp", "Session resumed successfully.");
-        _auth_success(conn);
+        _stream_negotiation_success(conn);
     } else if (strcmp(name, "failed") == 0) {
         name = NULL;
+        conn->sm_state->sm_enabled = 0;
 
         failed_cause =
             xmpp_stanza_get_child_by_ns(stanza, XMPP_NS_STANZAS_IETF);
         if (!failed_cause)
-            goto LBL_ERR;
+            goto err_sm;
 
         cause = xmpp_stanza_get_name(failed_cause);
         if (!cause)
-            goto LBL_ERR;
+            goto err_sm;
 
-        if (!strcmp(cause, "item-not-found") ||
-            !strcmp(cause, "feature-not-implemented")) {
+        if (!strcmp(cause, "item-not-found")) {
             if (conn->sm_state->resume) {
-                conn->sm_state->resume = 0;
-                conn->sm_state->can_resume = 0;
-                /* remember that the server reports having support
-                 * for resumption, but actually it doesn't ...
+                /* It's no error if there's no `h` attribute included
+                 * but if there is, it gives a hint at what the server
+                 * already received.
                  */
-                conn->sm_state->dont_request_resume =
-                    !strcmp(cause, "feature-not-implemented");
-                strophe_free(conn->ctx, conn->sm_state->previd);
-                conn->sm_state->previd = NULL;
-                strophe_free(conn->ctx, conn->sm_state->bound_jid);
-                conn->sm_state->bound_jid = NULL;
-                _do_bind(conn, conn->sm_state->bind);
-                conn->sm_state->bind = NULL;
+                if (!_get_h_attribute(stanza, &ul_h)) {
+                    /* In cases there's no `h` included, drop all elements. */
+                    ul_h = (unsigned long)-1;
+                }
+                _sm_queue_cleanup(conn, ul_h);
             }
+        } else if (!strcmp(cause, "feature-not-implemented")) {
+            conn->sm_state->resume = 0;
+            conn->sm_state->can_resume = 0;
+            /* remember that the server reports having support
+             * for resumption, but actually it doesn't ...
+             */
+            conn->sm_state->dont_request_resume = 1;
         }
-        conn->sm_state->sm_handled_nr = 0;
+        bind = conn->sm_state->bind;
+        conn->sm_state->bind = NULL;
+        reset_sm_state(conn->sm_state);
+        _do_bind(conn, bind);
     } else {
         /* unknown stanza received */
         name = NULL;
     }
 
-LBL_ERR:
+err_sm:
     if (!name) {
         char *err = "Couldn't convert stanza to text!";
         char *buf;
@@ -1277,7 +1491,8 @@ LBL_ERR:
                      buf);
         if (buf != err)
             strophe_free(conn->ctx, buf);
-        conn->sm_state->sm_enabled = 0;
+        /* Don't disable for <failure> cases, they're no hard errors */
+        conn->sm_state->sm_enabled = bind != NULL;
     }
     return 0;
 }
@@ -1310,7 +1525,7 @@ _handle_legacy(xmpp_conn_t *conn, xmpp_stanza_t *stanza, void *userdata)
         /* auth succeeded */
         strophe_debug(conn->ctx, "xmpp", "Legacy auth succeeded.");
 
-        _auth_success(conn);
+        _stream_negotiation_success(conn);
     } else {
         strophe_error(conn->ctx, "xmpp",
                       "Server sent us a legacy authentication "
@@ -1500,7 +1715,7 @@ int _handle_component_hs_response(xmpp_conn_t *conn,
         xmpp_disconnect(conn);
         return XMPP_EINT;
     } else {
-        _auth_success(conn);
+        _stream_negotiation_success(conn);
     }
 
     /* We don't need this handler anymore, return 0 so it can be deleted
@@ -1522,8 +1737,8 @@ int _handle_missing_handshake(xmpp_conn_t *conn, void *userdata)
 void auth_handle_open_raw(xmpp_conn_t *conn)
 {
     handler_reset_timed(conn, 0);
-    /* user handlers are not called before authentication is completed. */
-    _auth_success(conn);
+    /* user handlers are not called before stream negotiation has completed. */
+    _stream_negotiation_success(conn);
 }
 
 void auth_handle_open_stub(xmpp_conn_t *conn)
