@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: MIT OR GPL-3.0-only */
 /* conn.c
 ** strophe XMPP client library -- connection object functions
 **
@@ -6,7 +7,7 @@
 **  This software is provided AS-IS with no warranty, either express
 **  or implied.
 **
-**  This program is dual licensed under the MIT and GPLv3 licenses.
+**  This program is dual licensed under the MIT or GPLv3 licenses.
 */
 
 /** @file
@@ -29,7 +30,6 @@
 #include "common.h"
 #include "util.h"
 #include "parser.h"
-#include "resolver.h"
 
 #ifndef DEFAULT_SEND_QUEUE_MAX
 /** @def DEFAULT_SEND_QUEUE_MAX
@@ -78,6 +78,7 @@
 #define KEEPALIVE_COUNT 3
 #endif
 
+static int _is_connected(xmpp_conn_t *conn, xmpp_send_queue_owner_t owner);
 static int _disconnect_cleanup(xmpp_conn_t *conn, void *userdata);
 static void _reset_sm_state_for_reconnect(xmpp_conn_t *conn);
 static char *_conn_build_stream_tag(xmpp_conn_t *conn,
@@ -103,8 +104,6 @@ static unsigned short _conn_default_port(xmpp_conn_t *conn,
 static void _conn_reset(xmpp_conn_t *conn);
 static int _conn_connect(xmpp_conn_t *conn,
                          const char *domain,
-                         const char *host,
-                         unsigned short port,
                          xmpp_conn_type_t type,
                          xmpp_conn_handler callback,
                          void *userdata);
@@ -123,8 +122,6 @@ void xmpp_send_error(xmpp_conn_t *conn, xmpp_error_type_t type, char *text)
     xmpp_stanza_t *error = xmpp_error_new(conn->ctx, type, text);
 
     send_stanza(conn, error, XMPP_QUEUE_STROPHE);
-
-    xmpp_stanza_release(error);
 }
 
 /** Create a new Strophe connection object.
@@ -145,27 +142,19 @@ xmpp_conn_t *xmpp_conn_new(xmpp_ctx_t *ctx)
 
     conn = strophe_alloc(ctx, sizeof(xmpp_conn_t));
     if (conn != NULL) {
+        memset(conn, 0, sizeof(xmpp_conn_t));
         conn->ctx = ctx;
 
         conn->type = XMPP_UNKNOWN;
         conn->state = XMPP_STATE_DISCONNECTED;
 
-        conn->sock = -1;
+        conn->sock = INVALID_SOCKET;
         conn->ka_timeout = KEEPALIVE_TIMEOUT;
         conn->ka_interval = KEEPALIVE_INTERVAL;
         conn->ka_count = KEEPALIVE_COUNT;
-        conn->tls = NULL;
-        conn->timeout_stamp = 0;
-        conn->error = 0;
-        conn->stream_error = NULL;
 
         /* default send parameters */
-        conn->blocking_send = 0;
         conn->send_queue_max = DEFAULT_SEND_QUEUE_MAX;
-        conn->send_queue_len = 0;
-        conn->send_queue_user_len = 0;
-        conn->send_queue_head = NULL;
-        conn->send_queue_tail = NULL;
 
         /* default timeouts */
         conn->connect_timeout = CONNECT_TIMEOUT;
@@ -175,49 +164,14 @@ xmpp_conn_t *xmpp_conn_new(xmpp_ctx_t *ctx)
             strophe_free(conn->ctx, conn);
             return NULL;
         }
-        conn->domain = NULL;
-        conn->jid = NULL;
-        conn->pass = NULL;
-        conn->stream_id = NULL;
-        conn->bound_jid = NULL;
-
-        conn->is_raw = 0;
-        conn->tls_support = 0;
-        conn->tls_disabled = 0;
-        conn->tls_mandatory = 0;
-        conn->tls_legacy_ssl = 0;
-        conn->tls_trust = 0;
-        conn->tls_failed = 0;
-        conn->tls_cafile = NULL;
-        conn->tls_capath = NULL;
-        conn->tls_client_cert = NULL;
-        conn->tls_client_key = NULL;
-        conn->sasl_support = 0;
-        conn->auth_legacy_enabled = 0;
-        conn->secured = 0;
-        conn->certfail_handler = NULL;
-        conn->password_callback = NULL;
-        conn->password_callback_userdata = NULL;
         tls_clear_password_cache(conn);
         conn->password_retries = 1;
 
-        conn->bind_required = 0;
-        conn->session_required = 0;
-        conn->sm_state = NULL;
-
         conn->parser =
             parser_new(conn->ctx, _handle_stream_start, _handle_stream_end,
                        _handle_stream_stanza, conn);
-        conn->reset_parser = 0;
-
-        conn->authenticated = 0;
-        conn->conn_handler = NULL;
-        conn->userdata = NULL;
-        conn->timed_handlers = NULL;
         /* we own (and will free) the hash values */
         conn->id_handlers = hash_new(conn->ctx, 32, NULL);
-        conn->handlers = NULL;
-        conn->sockopt_cb = NULL;
 
         /* give the caller a reference to connection */
         conn->ref = 1;
@@ -404,6 +358,7 @@ int xmpp_conn_release(xmpp_conn_t *conn)
         if (conn->sm_state)
             xmpp_free_sm_state(conn->sm_state);
         tls_clear_password_cache(conn);
+        sock_free(conn->xsock);
         strophe_free(ctx, conn);
         released = 1;
     }
@@ -483,6 +438,8 @@ void xmpp_conn_set_certfail_handler(xmpp_conn_t *const conn,
  */
 void xmpp_conn_set_cafile(xmpp_conn_t *const conn, const char *path)
 {
+    if (conn->tls_cafile)
+        strophe_free(conn->ctx, conn->tls_cafile);
     conn->tls_cafile = strophe_strdup(conn->ctx, path);
 }
 
@@ -495,6 +452,8 @@ void xmpp_conn_set_cafile(xmpp_conn_t *const conn, const char *path)
  */
 void xmpp_conn_set_capath(xmpp_conn_t *const conn, const char *path)
 {
+    if (conn->tls_capath)
+        strophe_free(conn->ctx, conn->tls_capath);
     conn->tls_capath = strophe_strdup(conn->ctx, path);
 }
 
@@ -703,12 +662,7 @@ int xmpp_connect_client(xmpp_conn_t *conn,
                         xmpp_conn_handler callback,
                         void *userdata)
 {
-    resolver_srv_rr_t *srv_rr_list = NULL;
-    resolver_srv_rr_t *rr;
     char *domain;
-    const char *host = NULL;
-    unsigned short port = 0;
-    int found = XMPP_DOMAIN_NOT_FOUND;
     int rc;
 
     if (!conn->jid && (conn->tls_client_cert || conn->tls_client_key)) {
@@ -737,48 +691,35 @@ int xmpp_connect_client(xmpp_conn_t *conn,
     if (!conn->sm_state) {
         conn->sm_state = strophe_alloc(conn->ctx, sizeof(*conn->sm_state));
         if (!conn->sm_state)
-            return XMPP_EMEM;
+            goto err_mem;
         memset(conn->sm_state, 0, sizeof(*conn->sm_state));
         conn->sm_state->ctx = conn->ctx;
     }
 
-    if (altdomain != NULL) {
-        strophe_debug(conn->ctx, "xmpp", "Connecting via altdomain.");
-        host = altdomain;
-        port = altport ? altport : _conn_default_port(conn, XMPP_CLIENT);
-        found = XMPP_DOMAIN_ALTDOMAIN;
+    if (altdomain != NULL)
+        strophe_debug(conn->ctx, "conn", "Connecting via altdomain.");
 
+    if (conn->tls_legacy_ssl && !altdomain) {
         /* SSL tunneled connection on 5223 port is legacy and doesn't
          * have an SRV record. */
-    } else if (!conn->tls_legacy_ssl) {
-        found = resolver_srv_lookup(conn->ctx, "xmpp-client", "tcp", domain,
-                                    &srv_rr_list);
+        altdomain = domain;
     }
+    altport = altport ? altport : _conn_default_port(conn, XMPP_CLIENT);
 
-    if (XMPP_DOMAIN_NOT_FOUND == found) {
-        strophe_debug(conn->ctx, "xmpp",
-                      "SRV lookup failed, "
-                      "connecting via domain.");
-        host = domain;
-        port = altport ? altport : _conn_default_port(conn, XMPP_CLIENT);
-        found = XMPP_DOMAIN_ALTDOMAIN;
-    }
-
-    rr = srv_rr_list;
-    do {
-        if (XMPP_DOMAIN_FOUND == found && rr != NULL) {
-            host = rr->target;
-            port = rr->port;
-            rr = rr->next;
-        }
-        rc = _conn_connect(conn, domain, host, port, XMPP_CLIENT, callback,
-                           userdata);
-    } while (rc != 0 && rr != NULL);
+    if (conn->xsock)
+        sock_free(conn->xsock);
+    conn->xsock = sock_new(conn, domain, altdomain, altport);
+    if (!conn->xsock)
+        goto err_mem;
 
+    rc = _conn_connect(conn, domain, XMPP_CLIENT, callback, userdata);
     strophe_free(conn->ctx, domain);
-    resolver_srv_free(conn->ctx, srv_rr_list);
 
     return rc;
+
+err_mem:
+    strophe_free(conn->ctx, domain);
+    return XMPP_EMEM;
 }
 
 /** Initiate a component connection to server.
@@ -815,7 +756,8 @@ int xmpp_connect_component(xmpp_conn_t *conn,
         return XMPP_EINVOP;
 
     /* XEP-0114 does not support TLS */
-    xmpp_conn_disable_tls(conn);
+    (void)xmpp_conn_set_flags(conn, xmpp_conn_get_flags(conn) |
+                                        XMPP_CONN_FLAG_DISABLE_TLS);
     if (!conn->tls_disabled) {
         strophe_error(conn->ctx, "conn",
                       "Failed to disable TLS. "
@@ -824,10 +766,15 @@ int xmpp_connect_component(xmpp_conn_t *conn,
     }
 
     port = port ? port : _conn_default_port(conn, XMPP_COMPONENT);
+    if (conn->xsock)
+        sock_free(conn->xsock);
+    conn->xsock = sock_new(conn, NULL, server, port);
+    if (!conn->xsock)
+        return XMPP_EMEM;
+
     /* JID serves as an identifier here and will be used as "to" attribute
        of the stream */
-    return _conn_connect(conn, conn->jid, server, port, XMPP_COMPONENT,
-                         callback, userdata);
+    return _conn_connect(conn, conn->jid, XMPP_COMPONENT, callback, userdata);
 }
 
 /** Initiate a raw connection to the XMPP server.
@@ -874,9 +821,10 @@ void conn_established(xmpp_conn_t *conn)
 
     if (conn->is_raw) {
         handler_reset_timed(conn, 0);
-        /* we skip authentication for a "raw" connection, but the event loop
-           ignores user's handlers when conn->authenticated is not set. */
-        conn->authenticated = 1;
+        /* we skip all the mandatory steps of the stream negotiation for a "raw"
+           connection, but the event loop ignores user's handlers when
+           conn->stream_negotiation_completed is not set. */
+        conn->stream_negotiation_completed = 1;
         conn->conn_handler(conn, XMPP_CONN_RAW_CONNECT, 0, NULL,
                            conn->userdata);
     } else {
@@ -970,12 +918,14 @@ void conn_disconnect(xmpp_conn_t *conn)
 {
     strophe_debug(conn->ctx, "xmpp", "Closing socket.");
     conn->state = XMPP_STATE_DISCONNECTED;
+    conn->stream_negotiation_completed = 0;
     if (conn->tls) {
         tls_stop(conn->tls);
         tls_free(conn->tls);
         conn->tls = NULL;
     }
-    sock_close(conn->sock);
+    if (conn->sock != INVALID_SOCKET)
+        sock_close(conn->sock);
     _reset_sm_state_for_reconnect(conn);
 
     /* fire off connection handler */
@@ -1037,7 +987,7 @@ void xmpp_send_raw_string(xmpp_conn_t *conn, const char *fmt, ...)
 {
     va_list ap;
 
-    if (conn->state != XMPP_STATE_CONNECTED)
+    if (!_is_connected(conn, XMPP_QUEUE_USER))
         return;
 
     va_start(ap, fmt);
@@ -1073,7 +1023,7 @@ void xmpp_send_raw(xmpp_conn_t *conn, const char *data, size_t len)
  */
 void xmpp_send(xmpp_conn_t *conn, xmpp_stanza_t *stanza)
 {
-    send_stanza(conn, stanza, XMPP_QUEUE_USER);
+    send_stanza(conn, xmpp_stanza_clone(stanza), XMPP_QUEUE_USER);
 }
 
 /** Send the opening &lt;stream:stream&gt; tag to the server.
@@ -1112,6 +1062,23 @@ void conn_open_stream(xmpp_conn_t *conn)
         strophe_free(conn->ctx, from);
 }
 
+int conn_interface_write(struct conn_interface *intf,
+                         const void *buff,
+                         size_t len)
+{
+    int ret = intf->write(intf, buff, len);
+    if (ret < 0 && !intf->error_is_recoverable(intf, intf->get_error(intf))) {
+        intf->conn->error = intf->get_error(intf);
+    }
+    return ret;
+}
+
+int conn_int_nop(struct conn_interface *intf)
+{
+    UNUSED(intf);
+    return 0;
+}
+
 int conn_tls_start(xmpp_conn_t *conn)
 {
     int rc;
@@ -1125,11 +1092,13 @@ int conn_tls_start(xmpp_conn_t *conn)
     }
 
     if (conn->tls != NULL) {
+        conn->intf = tls_intf;
+        conn->intf.conn = conn;
         if (tls_start(conn->tls)) {
             conn->secured = 1;
         } else {
             rc = XMPP_EINT;
-            conn->error = tls_error(conn->tls);
+            conn->error = tls_error(&conn->intf);
             tls_free(conn->tls);
             conn->tls = NULL;
             conn->tls_failed = 1;
@@ -1156,11 +1125,15 @@ long xmpp_conn_get_flags(const xmpp_conn_t *conn)
 {
     long flags;
 
-    flags = XMPP_CONN_FLAG_DISABLE_TLS * conn->tls_disabled |
-            XMPP_CONN_FLAG_MANDATORY_TLS * conn->tls_mandatory |
-            XMPP_CONN_FLAG_LEGACY_SSL * conn->tls_legacy_ssl |
-            XMPP_CONN_FLAG_TRUST_TLS * conn->tls_trust |
-            XMPP_CONN_FLAG_LEGACY_AUTH * conn->auth_legacy_enabled;
+    flags =
+        XMPP_CONN_FLAG_DISABLE_TLS * conn->tls_disabled |
+        XMPP_CONN_FLAG_MANDATORY_TLS * conn->tls_mandatory |
+        XMPP_CONN_FLAG_LEGACY_SSL * conn->tls_legacy_ssl |
+        XMPP_CONN_FLAG_TRUST_TLS * conn->tls_trust |
+        XMPP_CONN_FLAG_DISABLE_SM * conn->sm_disable |
+        XMPP_CONN_FLAG_ENABLE_COMPRESSION * conn->compression.allowed |
+        XMPP_CONN_FLAG_COMPRESSION_DONT_RESET * conn->compression.dont_reset |
+        XMPP_CONN_FLAG_LEGACY_AUTH * conn->auth_legacy_enabled;
 
     return flags;
 }
@@ -1180,6 +1153,8 @@ long xmpp_conn_get_flags(const xmpp_conn_t *conn)
  *    - XMPP_CONN_FLAG_TRUST_TLS
  *    - XMPP_CONN_FLAG_LEGACY_AUTH
  *    - XMPP_CONN_FLAG_DISABLE_SM
+ *    - XMPP_CONN_FLAG_ENABLE_COMPRESSION
+ *    - XMPP_CONN_FLAG_COMPRESSION_DONT_RESET
  *
  *  @param conn a Strophe connection object
  *  @param flags ORed connection flags
@@ -1209,29 +1184,23 @@ int xmpp_conn_set_flags(xmpp_conn_t *conn, long flags)
     conn->tls_trust = (flags & XMPP_CONN_FLAG_TRUST_TLS) ? 1 : 0;
     conn->auth_legacy_enabled = (flags & XMPP_CONN_FLAG_LEGACY_AUTH) ? 1 : 0;
     conn->sm_disable = (flags & XMPP_CONN_FLAG_DISABLE_SM) ? 1 : 0;
+    conn->compression.allowed =
+        (flags & XMPP_CONN_FLAG_ENABLE_COMPRESSION) ? 1 : 0;
+    conn->compression.dont_reset =
+        (flags & XMPP_CONN_FLAG_COMPRESSION_DONT_RESET) ? 1 : 0;
+    flags &= ~(XMPP_CONN_FLAG_DISABLE_TLS | XMPP_CONN_FLAG_MANDATORY_TLS |
+               XMPP_CONN_FLAG_LEGACY_SSL | XMPP_CONN_FLAG_TRUST_TLS |
+               XMPP_CONN_FLAG_LEGACY_AUTH | XMPP_CONN_FLAG_DISABLE_SM |
+               XMPP_CONN_FLAG_ENABLE_COMPRESSION |
+               XMPP_CONN_FLAG_COMPRESSION_DONT_RESET);
+    if (flags) {
+        strophe_error(conn->ctx, "conn", "Flags 0x%04lx unknown", flags);
+        return XMPP_EINVOP;
+    }
 
     return 0;
 }
 
-/** Disable TLS for this connection, called by users of the library.
- *  Occasionally a server will be misconfigured to send the starttls
- *  feature, but will not support the handshake.
- *
- *  @param conn a Strophe connection object
- *
- *  @note this function is deprecated
- *  @see xmpp_conn_set_flags()
- *
- *  @ingroup Connections
- */
-void xmpp_conn_disable_tls(xmpp_conn_t *conn)
-{
-    long flags = xmpp_conn_get_flags(conn);
-
-    flags |= XMPP_CONN_FLAG_DISABLE_TLS;
-    (void)xmpp_conn_set_flags(conn, flags);
-}
-
 /** Return whether TLS session is established or not.
  *
  *  @return TRUE if TLS session is established and FALSE otherwise
@@ -1250,17 +1219,26 @@ int xmpp_conn_is_secured(xmpp_conn_t *conn)
  */
 int xmpp_conn_is_connecting(xmpp_conn_t *conn)
 {
-    return conn->state == XMPP_STATE_CONNECTING;
+    return conn->state == XMPP_STATE_CONNECTING ||
+           (conn->state == XMPP_STATE_CONNECTED &&
+            conn->stream_negotiation_completed == 0);
+}
+
+static int _is_connected(xmpp_conn_t *conn, xmpp_send_queue_owner_t owner)
+{
+    return conn->state == XMPP_STATE_CONNECTED &&
+           (owner != XMPP_QUEUE_USER ||
+            conn->stream_negotiation_completed == 1);
 }
 
 /**
- *  @return TRUE if connection is in connected state and FALSE otherwise
+ *  @return TRUE if connection is established and FALSE otherwise
  *
  *  @ingroup Connections
  */
 int xmpp_conn_is_connected(xmpp_conn_t *conn)
 {
-    return conn->state == XMPP_STATE_CONNECTED;
+    return _is_connected(conn, XMPP_QUEUE_USER);
 }
 
 /**
@@ -1324,7 +1302,7 @@ static void _reset_sm_state_for_reconnect(xmpp_conn_t *conn)
         s->id = NULL;
     }
 
-    s->sm_enabled = s->sm_support = s->resume = 0;
+    s->r_sent = s->sm_enabled = s->sm_support = s->resume = 0;
 
     if (s->bind) {
         xmpp_stanza_release(s->bind);
@@ -1344,8 +1322,11 @@ static void _reset_sm_state_for_reconnect(xmpp_conn_t *conn)
 int xmpp_conn_set_sm_state(xmpp_conn_t *conn, xmpp_sm_state_t *sm_state)
 {
     /* We can only set the SM state when we're disconnected */
-    if (conn->state != XMPP_STATE_DISCONNECTED)
+    if (conn->state != XMPP_STATE_DISCONNECTED) {
+        strophe_error(conn->ctx, "conn",
+                      "SM state can only be set the when we're disconnected");
         return XMPP_EINVOP;
+    }
 
     if (conn->sm_state) {
         strophe_error(conn->ctx, "conn", "SM state is already set!");
@@ -1364,6 +1345,21 @@ int xmpp_conn_set_sm_state(xmpp_conn_t *conn, xmpp_sm_state_t *sm_state)
     return XMPP_EOK;
 }
 
+void reset_sm_state(xmpp_sm_state_t *sm_state)
+{
+    xmpp_ctx_t *ctx = sm_state->ctx;
+
+    strophe_free_and_null(ctx, sm_state->id);
+    strophe_free_and_null(ctx, sm_state->previd);
+    strophe_free_and_null(ctx, sm_state->bound_jid);
+    if (sm_state->bind)
+        xmpp_stanza_release(sm_state->bind);
+    sm_state->bind = NULL;
+    sm_state->sm_handled_nr = 0;
+    sm_state->sm_sent_nr = 0;
+    sm_state->r_sent = 0;
+}
+
 /**  c.f. \ref xmpp_conn_get_sm_state for usage documentation
  *
  *  @param sm_state   A Stream Management state returned from a call to
@@ -1373,8 +1369,8 @@ int xmpp_conn_set_sm_state(xmpp_conn_t *conn, xmpp_sm_state_t *sm_state)
  */
 void xmpp_free_sm_state(xmpp_sm_state_t *sm_state)
 {
-    xmpp_ctx_t *ctx;
     xmpp_send_queue_t *smq;
+    xmpp_ctx_t *ctx;
 
     if (!sm_state || !sm_state->ctx)
         return;
@@ -1385,15 +1381,7 @@ void xmpp_free_sm_state(xmpp_sm_state_t *sm_state)
         strophe_free(ctx, queue_element_free(ctx, smq));
     }
 
-    if (sm_state->bind)
-        xmpp_stanza_release(sm_state->bind);
-    if (sm_state->id)
-        strophe_free(ctx, sm_state->id);
-    if (sm_state->previd)
-        strophe_free(ctx, sm_state->previd);
-    if (sm_state->bound_jid)
-        strophe_free(ctx, sm_state->bound_jid);
-    memset(sm_state, 0, sizeof(*sm_state));
+    reset_sm_state(sm_state);
     strophe_free(ctx, sm_state);
 }
 
@@ -1738,7 +1726,6 @@ static void _conn_sm_handle_stanza(xmpp_conn_t *const conn,
             strophe_snprintf(h, sizeof(h), "%u", conn->sm_state->sm_handled_nr);
             xmpp_stanza_set_attribute(a, "h", h);
             send_stanza(conn, a, XMPP_QUEUE_SM_STROPHE);
-            xmpp_stanza_release(a);
         } else if (strcmp(name, "a") == 0) {
             attr_h = xmpp_stanza_get_attribute(stanza, "h");
             if (!attr_h) {
@@ -1758,10 +1745,12 @@ static void _conn_sm_handle_stanza(xmpp_conn_t *const conn,
             while (conn->sm_state->sm_queue.head &&
                    conn->sm_state->sm_queue.head->sm_h < ul_h) {
                 e = pop_queue_front(&conn->sm_state->sm_queue);
-                strophe_debug_verbose(2, conn->ctx, "conn", "SM_Q_DROP: %p", e);
+                strophe_debug_verbose(2, conn->ctx, "conn",
+                                      "SM_Q_DROP: %p, h=%lu", e, e->sm_h);
                 c = queue_element_free(conn->ctx, e);
                 strophe_free(conn->ctx, c);
             }
+            conn->sm_state->r_sent = 0;
         }
     }
 }
@@ -1800,6 +1789,11 @@ static void _conn_reset(xmpp_conn_t *conn)
         return;
     }
 
+    compression_free(conn);
+
+    conn->intf = sock_intf;
+    conn->intf.conn = conn;
+
     /* free queued */
     sq = conn->send_queue_head;
     while (sq) {
@@ -1814,22 +1808,14 @@ static void _conn_reset(xmpp_conn_t *conn)
 
     if (conn->stream_error) {
         xmpp_stanza_release(conn->stream_error->stanza);
-        if (conn->stream_error->text)
-            strophe_free(ctx, conn->stream_error->text);
-        strophe_free(ctx, conn->stream_error);
-        conn->stream_error = NULL;
+        strophe_free_and_null(ctx, conn->stream_error->text);
+        strophe_free_and_null(ctx, conn->stream_error);
     }
 
-    if (conn->domain)
-        strophe_free(ctx, conn->domain);
-    if (conn->bound_jid)
-        strophe_free(ctx, conn->bound_jid);
-    if (conn->stream_id)
-        strophe_free(ctx, conn->stream_id);
-    conn->domain = NULL;
-    conn->bound_jid = NULL;
-    conn->stream_id = NULL;
-    conn->authenticated = 0;
+    strophe_free_and_null(ctx, conn->domain);
+    strophe_free_and_null(ctx, conn->bound_jid);
+    strophe_free_and_null(ctx, conn->stream_id);
+    conn->stream_negotiation_completed = 0;
     conn->secured = 0;
     conn->tls_failed = 0;
     conn->error = 0;
@@ -1844,8 +1830,6 @@ static void _conn_reset(xmpp_conn_t *conn)
 
 static int _conn_connect(xmpp_conn_t *conn,
                          const char *domain,
-                         const char *host,
-                         unsigned short port,
                          xmpp_conn_type_t type,
                          xmpp_conn_handler callback,
                          void *userdata)
@@ -1856,8 +1840,6 @@ static int _conn_connect(xmpp_conn_t *conn,
         return XMPP_EINVOP;
     if (type != XMPP_CLIENT && type != XMPP_COMPONENT)
         return XMPP_EINVOP;
-    if (host == NULL || port == 0)
-        return XMPP_EINT;
 
     _conn_reset(conn);
 
@@ -1866,10 +1848,8 @@ static int _conn_connect(xmpp_conn_t *conn,
     if (!conn->domain)
         return XMPP_EMEM;
 
-    conn->sock = sock_connect(conn, host, port);
-    strophe_debug(conn->ctx, "xmpp", "sock_connect() to %s:%u returned %d",
-                  host, port, conn->sock);
-    if (conn->sock == -1)
+    conn->sock = sock_connect(conn->xsock);
+    if (conn->sock == INVALID_SOCKET)
         return XMPP_EINT;
 
     /* setup handler */
@@ -1888,7 +1868,6 @@ static int _conn_connect(xmpp_conn_t *conn,
 
     conn->state = XMPP_STATE_CONNECTING;
     conn->timeout_stamp = time_stamp();
-    strophe_debug(conn->ctx, "xmpp", "Attempting to connect to %s", host);
 
     return 0;
 }
@@ -1923,7 +1902,7 @@ static void _send_valist(xmpp_conn_t *conn,
     char buf[1024]; /* small buffer for common case */
     char *bigbuf;
 
-    if (conn->state != XMPP_STATE_CONNECTED)
+    if (!_is_connected(conn, owner))
         return;
 
     va_copy(apdup, ap);
@@ -1971,15 +1950,17 @@ void send_stanza(xmpp_conn_t *conn,
     char *buf = NULL;
     size_t len;
 
-    if (conn->state != XMPP_STATE_CONNECTED)
-        return;
+    if (!_is_connected(conn, owner))
+        goto out;
 
     if (xmpp_stanza_to_text(stanza, &buf, &len) != 0) {
         strophe_error(conn->ctx, "conn", "Failed to stanza_to_text");
-        return;
+        goto out;
     }
 
     _send_raw(conn, buf, len, owner, NULL);
+out:
+    xmpp_stanza_release(stanza);
 }
 
 void add_queue_back(xmpp_queue_t *queue, xmpp_send_queue_t *item)
@@ -1996,6 +1977,11 @@ void add_queue_back(xmpp_queue_t *queue, xmpp_send_queue_t *item)
     }
 }
 
+xmpp_send_queue_t *peek_queue_front(xmpp_queue_t *queue)
+{
+    return queue->head;
+}
+
 xmpp_send_queue_t *pop_queue_front(xmpp_queue_t *queue)
 {
     xmpp_send_queue_t *ret = queue->head;
@@ -2051,8 +2037,10 @@ static int _send_raw(xmpp_conn_t *conn,
         conn->send_queue_user_len++;
     strophe_debug_verbose(3, conn->ctx, "conn", "QUEUED: %s", data);
     strophe_debug_verbose(1, conn->ctx, "conn", "Q_ADD: %p", item);
-    if (!(owner & XMPP_QUEUE_SM) && conn->sm_state->sm_enabled) {
+    if (!(owner & XMPP_QUEUE_SM) && conn->sm_state->sm_enabled &&
+        !conn->sm_state->r_sent) {
         send_raw(conn, req_ack, strlen(req_ack), XMPP_QUEUE_SM_STROPHE, item);
+        conn->sm_state->r_sent = 1;
     }
     return XMPP_EOK;
 }
