@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: MIT OR GPL-3.0-only */
 /* tls_openssl.c
 ** strophe XMPP client library -- TLS abstraction openssl impl.
 **
@@ -6,7 +7,7 @@
 **  This software is provided AS-IS with no warranty, either express
 **  or implied.
 **
-**  This program is dual licensed under the MIT and GPLv3 licenses.
+**  This program is dual licensed under the MIT or GPLv3 licenses.
 */
 
 /** @file
@@ -106,6 +107,8 @@ struct _tls {
     SSL_CTX *ssl_ctx;
     SSL *ssl;
     X509 *client_cert;
+    void *channel_binding_data;
+    size_t channel_binding_size;
     int lasterror;
 };
 
@@ -280,9 +283,9 @@ void tls_shutdown(void)
 #endif
 }
 
-int tls_error(tls_t *tls)
+int tls_error(struct conn_interface *intf)
 {
-    return tls->lasterror;
+    return intf->conn->tls->lasterror;
 }
 
 /** Search through the SubjectAlternativeNames and return the next
@@ -411,8 +414,10 @@ _get_alg(const xmpp_ctx_t *ctx, X509 *err_cert, xmpp_cert_element_t el)
         alg_nid = OBJ_obj2nid(err_cert->sig_alg->algorithm);
 #else
         const X509_ALGOR *palg;
+        const ASN1_OBJECT *obj;
         X509_get0_signature(NULL, &palg, err_cert);
-        alg_nid = OBJ_obj2nid(palg->algorithm);
+        X509_ALGOR_get0(&obj, NULL, NULL, palg);
+        alg_nid = OBJ_obj2nid(obj);
 #endif
     } break;
     default:
@@ -712,6 +717,7 @@ err:
 
 void tls_free(tls_t *tls)
 {
+    strophe_free(tls->ctx, tls->channel_binding_data);
     SSL_free(tls->ssl);
     X509_free(tls->client_cert);
     SSL_CTX_free(tls->ssl_ctx);
@@ -738,6 +744,65 @@ int tls_set_credentials(tls_t *tls, const char *cafilename)
     return -1;
 }
 
+int tls_init_channel_binding(tls_t *tls,
+                             const char **binding_prefix,
+                             size_t *binding_prefix_len)
+{
+    const char *label = NULL;
+    size_t labellen = 0;
+
+    switch (SSL_version(tls->ssl)) {
+    case SSL3_VERSION:
+        *binding_prefix = "tls-unique";
+        *binding_prefix_len = strlen("tls-unique");
+        tls->channel_binding_size = 36;
+        break;
+    case TLS1_VERSION:
+    case TLS1_1_VERSION:
+    case TLS1_2_VERSION:
+        *binding_prefix = "tls-unique";
+        *binding_prefix_len = strlen("tls-unique");
+        tls->channel_binding_size = 12;
+        break;
+#ifdef TLS1_3_VERSION
+    case TLS1_3_VERSION:
+        label = "EXPORTER-Channel-Binding";
+        labellen = 24;
+        *binding_prefix = "tls-exporter";
+        *binding_prefix_len = strlen("tls-exporter");
+        tls->channel_binding_size = 32;
+        break;
+#endif
+    default:
+        strophe_error(tls->ctx, "tls", "Unsupported TLS Version: %s",
+                      SSL_get_version(tls->ssl));
+        return -1;
+    }
+
+    strophe_free_and_null(tls->ctx, tls->channel_binding_data);
+    tls->channel_binding_data =
+        strophe_alloc(tls->ctx, tls->channel_binding_size);
+    if (!tls->channel_binding_data)
+        return -1;
+
+    if (SSL_export_keying_material(tls->ssl, tls->channel_binding_data,
+                                   tls->channel_binding_size, label, labellen,
+                                   NULL, 0, 0) != 1) {
+        strophe_error(tls->ctx, "tls", "Could not get channel binding data");
+        return -1;
+    }
+    return 0;
+}
+
+const void *tls_get_channel_binding_data(tls_t *tls, size_t *size)
+{
+    if (!tls->channel_binding_data || !tls->channel_binding_size) {
+        strophe_error(tls->ctx, "tls", "No channel binding data available");
+    }
+    *size = tls->channel_binding_size;
+    return tls->channel_binding_data;
+}
+
 int tls_start(tls_t *tls)
 {
     int error;
@@ -750,7 +815,7 @@ int tls_start(tls_t *tls)
         ret = SSL_connect(tls->ssl);
         error = ret <= 0 ? SSL_get_error(tls->ssl, ret) : 0;
 
-        if (ret == -1 && tls_is_recoverable(error)) {
+        if (ret == -1 && tls_is_recoverable(NULL, error)) {
             /* wait for something to happen on the sock before looping back */
             _tls_sock_wait(tls, error);
             continue;
@@ -791,7 +856,7 @@ int tls_stop(tls_t *tls)
         ++retries;
         ret = SSL_shutdown(tls->ssl);
         error = ret < 0 ? SSL_get_error(tls->ssl, ret) : 0;
-        if (ret == 1 || !tls_is_recoverable(error) ||
+        if (ret == 1 || !tls_is_recoverable(NULL, error) ||
             retries >= TLS_SHUTDOWN_MAX_RETRIES) {
             break;
         }
@@ -810,21 +875,23 @@ int tls_stop(tls_t *tls)
     return ret <= 0 ? 0 : 1;
 }
 
-int tls_is_recoverable(int error)
+int tls_is_recoverable(struct conn_interface *intf, int error)
 {
+    UNUSED(intf);
     return (error == SSL_ERROR_NONE || error == SSL_ERROR_WANT_READ ||
             error == SSL_ERROR_WANT_WRITE || error == SSL_ERROR_WANT_CONNECT ||
             error == SSL_ERROR_WANT_ACCEPT);
 }
 
-int tls_pending(tls_t *tls)
+int tls_pending(struct conn_interface *intf)
 {
-    return SSL_pending(tls->ssl);
+    return SSL_pending(intf->conn->tls->ssl);
 }
 
-int tls_read(tls_t *tls, void *buff, size_t len)
+int tls_read(struct conn_interface *intf, void *buff, size_t len)
 {
     int ret;
+    tls_t *tls = intf->conn->tls;
 
     ret = SSL_read(tls->ssl, buff, len);
     _tls_set_error(tls, ret <= 0 ? SSL_get_error(tls->ssl, ret) : 0);
@@ -832,9 +899,10 @@ int tls_read(tls_t *tls, void *buff, size_t len)
     return ret;
 }
 
-int tls_write(tls_t *tls, const void *buff, size_t len)
+int tls_write(struct conn_interface *intf, const void *buff, size_t len)
 {
     int ret;
+    tls_t *tls = intf->conn->tls;
 
     ret = SSL_write(tls->ssl, buff, len);
     _tls_set_error(tls, ret <= 0 ? SSL_get_error(tls->ssl, ret) : 0);
@@ -842,9 +910,9 @@ int tls_write(tls_t *tls, const void *buff, size_t len)
     return ret;
 }
 
-int tls_clear_pending_write(tls_t *tls)
+int tls_clear_pending_write(struct conn_interface *intf)
 {
-    UNUSED(tls);
+    UNUSED(intf);
     return 0;
 }
 
@@ -882,7 +950,7 @@ static const char *_tls_error_str(int error, const char **tbl, size_t tbl_size)
 
 static void _tls_set_error(tls_t *tls, int error)
 {
-    if (error != 0 && !tls_is_recoverable(error)) {
+    if (error != 0 && !tls_is_recoverable(NULL, error)) {
         strophe_debug(tls->ctx, "tls", "error=%s(%d) errno=%d lasterror=%d",
                       TLS_ERROR_STR(error, tls_errors), error, errno,
                       tls->lasterror);
